################################## 
# Description: Receives records after subscribing to the MSG 
#      +  Validates the records
#      +  Insert the records on the corresponding db view
#      +  According to Grid Monitoring Probe Specification v0.91
# Author: Piotr Nyczyk(base code), Daniel Rodrigues
# Institute: CERN
# Created: January 08
# License: 
# $Id:$ 
################################## 

import os
import sys
import time
import threading
import cx_Oracle
from ConfigParser import ConfigParser,NoOptionError
from Queue import Queue
import stomp
import logging

config=None

class StatsCollector:
    def __init__(self):
        self.stats={}
        
    def setFlag(self,stat,flag,state=True):
        if not self.stats.has_key(stat):
            self.stats[stat]={}
        self.stats[stat][flag]=state
            
    def updateCounter(self,stat,increment=1):
        if not self.stats.has_key(stat):
            self.stats[stat]=0
        self.stats[stat]+=increment
    
    def setCounter(self,stat,value=0):
        self.stats[stat]=value

    def dumps(self):
        l=self.stats.keys()
        l.sort()
        s=""
        for k in l:
            s+="%s: %s\n"%(k,str(self.stats[k]))
        return s

stats=StatsCollector()
        
def read_config(configFile):
    global config
    config=ConfigParser()
    if os.path.exists(configFile) :
        config.read(configFile)
    else:
        logger.error( "Could not open config : %s"%configFile)
        raise OSError("Could not open config : %s"%configFile)

def log_config() :
    pass

def get_message_class(msgClass):
    try:
        attrs=map(str.lower,config.get('msg-classes',msgClass+'_dbattrs').split())
        table=config.get('msg-classes',msgClass+'_dbtable')
    except NoOptionError:
        logger.error('Unknow message class: %s!\n'%msgClass)
        sys.exit(1)
    return table,attrs

def create_insert_sql(table,attrs):
    colstr=','.join(attrs)
    valstr=','.join(map(lambda x: ':'+x,attrs))
    sql="insert into %s(%s) values(%s)"%(table,colstr,valstr)
    return sql

def validate_record(record,attrs,cursor):
    empty=True
    for attr in attrs:
        if attr not in record:
            record[attr]=None
        else:
            if len(record[attr])>1000:
                v=cursor.var(cx_Oracle.LONG_STRING,len(record[attr]))
                v.setvalue(0,record[attr])
                record[attr]=v
            empty=False
    for key in record.keys():
        if key not in attrs:
            del record[key]
    return not empty
            
def read_record(linesQueue):
    record={}
    key=''
    value=''
    multiline=False
    line=linesQueue.get()
    while line and line!='EOT':
        if multiline:
            value+='\n'+line
        elif ':' in line:
            key,value=map(str.strip, line.split(':',1))
            key=key.lower()
            if key=='detailsdata':
                multiline=True
            else:
                record[key]=value
        else:
            logger.error('Syntax error: '+line)
        line=linesQueue.get()
    if multiline:
        record[key]=value
    return record

class DBCommiter(threading.Thread):
    def __init__(self,conn,commitDelay):
        threading.Thread.__init__(self,name='DBCommiter')
        self.setDaemon(True)
        self.conn=conn
        self.lock=threading.Lock()
        self.commitDelay=commitDelay
        self.nRecords={}#0
        self.running=True

    def run(self):
        while self.running:
            time.sleep(self.commitDelay)
            self.acquire()
            self.conn.commit()
            self.release()
            #sys.stderr.write('%s: commited %d records\n'%(time.ctime(),self.nRecords))
            #self.nRecords=0

    def acquire(self):
        stats.setFlag('commiter_lock_waiting',threading.currentThread().getName(),True)
        self.lock.acquire()
        stats.setFlag('commiter_lock_waiting',threading.currentThread().getName(),False)
        stats.setFlag('commiter_lock_acquired',threading.currentThread().getName(),True)

    def release(self):
        self.lock.release()
        stats.setFlag('commiter_lock_acquired',threading.currentThread().getName(),False)
        

def format_record(record):
    multiline=None
    s=''

    for key,value in record.items():
        if not value:
            continue
        if key.lower() == 'detailsdata' :
            multiline='%s: %s\n'%(key,str(value))
        else:
            s+='%s: %s\n'%(key,str(value))
    if multiline:
        s+=multiline
    return s+'EOT\n#'

class BrokerLogHandler(logging.Handler):
    def __init__(self,destination,level=0):
        logging.Handler.__init__(self,level)
        self.destination=destination
    def emit(self,record):
        broker.send(self.destination,stomp.Message(record.getMessage()))

class DBInserter(threading.Thread):
    def __init__(self,commiter,msgClass,queueSize=1000):
        threading.Thread.__init__(self,name='DBInserter_'+msgClass)
        self.setDaemon(True)
        self.commiter=commiter
        self.cur=commiter.conn.cursor()
        self.table,self.attrs=get_message_class(msgClass)
        self.insert_sql=create_insert_sql(self.table,self.attrs)
        self.msgClass=msgClass
        self.queue=Queue(queueSize)
        logname='rejected_'+msgClass
        self.rejectedLog=logging.getLogger(logname)
        self.rejectedLog.setLevel(logging.DEBUG)
        form=logging.Formatter('# %(asctime)s %(message)s')
        logfile=config.get('logging','rejected-location')+'/'+logname+'.log'

# Does'n exist in Python 2.3
#        hdlr=logging.RotatingFileHandler(filename=logfile,
#                                         maxBytes=config.get('logging','rejected-size'),
#                                         backupCount=config.get('logging','rejected-count'))

        hdlr=logging.FileHandler(logfile)
        hdlr.setFormatter(form)
        self.rejectedLog.addHandler(hdlr)
        destination=config.get('logging','rejected-destination')+'.'+msgClass
        hdlr=BrokerLogHandler(destination)
        self.rejectedLog.addHandler(hdlr)
        

    def run(self):
        while True:
            stats.updateCounter(self.getName()+'_read_record',1)
            record=read_record(self.queue)
            stats.updateCounter(self.getName()+'_read_record',-1)
            try:
                self.commiter.acquire()
                if not validate_record(record,self.attrs,self.cur):
                    self.commiter.release()
                    logger.error("BAD RECORD\n")
                    self.rejectedLog.error('VALIDATION ERROR\n'+format_record(record))
                    continue
        
                self.cur.execute(self.insert_sql,record)
                self.commiter.nRecords.setdefault(self.msgClass,0)
                self.commiter.nRecords[self.msgClass]+=1
                self.commiter.release()
            except Exception, e:
                self.commiter.release()
                logger.exception("Got exception: %sWhile inserting record:\n %s\n\n"%(str(e),str(record)))
                estr=''.join(['# %s\n'%l for l in str(e).split('\n')])
                self.rejectedLog.error(estr+format_record(record))

    def receive(self,msg):
        stats.updateCounter(self.getName()+'_receive',1)            
        for line in msg.body.strip().split('\n'):
            self.queue.put(line)
        stats.updateCounter(self.getName()+'_receive',-1)            

commiter=None
inserters=[]
broker=None

def start():
    global commiter
    global inserters
    global broker
    
    db=cx_Oracle.connect(config.get('db','database'))
    
    broker_host=config.get('stomp','broker_host')
    broker_port=int(config.get('stomp','broker_port'))
    client_id=config.get('stomp','client_id')
    broker=stomp.Connection(broker_host,broker_port,client_id=client_id)

    commiter=DBCommiter(db,1)
    commiter.start()

    inserters={}
    for msgClass in config.get('msg-classes','classes').split():
        msgDest=config.get('msg-classes',msgClass+'_destination')
        msgInserter=DBInserter(commiter,msgClass)
        msgInserter.start()
        broker.subscribe(msgInserter.receive,msgDest,subscriptionName=msgClass)
        inserters[msgClass]=msgInserter
        logger.debug("Created DBInserter for message class %s using destination %s"%(msgClass,msgDest))


def stats_loop(refresh=5):
    while True:
        if commiter.nRecords:
            for msgClass in commiter.nRecords.keys() :
                if commiter.nRecords[msgClass] :
                    logger.info("Got %s records from %s" % (str(commiter.nRecords[msgClass]),msgClass))
                commiter.nRecords[msgClass]=0
        #print stats.dumps()
        #print inserters
        time.sleep(refresh)

def setup_logging(logFile, runAsDaemon, verbose):
    global logger

    logger=logging.getLogger('main')
    if verbose:
        logger.setLevel(logging.DEBUG)
    else :
        logger.setLevel(logging.INFO)

# Does'n exist in Python 2.3
#    hdlr=logging.RotatingFileHandler(filename=logfile,
#                                     maxBytes=config.get('logging','log-size'),
#                                     backupCount=config.get('logging','log-count'))

    format=logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    fileHdlr=logging.FileHandler(logFile)
    fileHdlr.setFormatter(format)
    logger.addHandler(fileHdlr)
    if not runAsDaemon: # Also log to STDOUT
        stdoutHdlr=logging.StreamHandler()
        stdoutHdlr.setFormatter(format)
        logger.addHandler(stdoutHdlr)


def daemonize(pidFile):
    # do the UNIX double-fork magic, see Stevens' "Advanced 
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try: 
        pid = os.fork() 
        if pid > 0:
            # exit first parent
            sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)

   # decouple from parent environment
    os.chdir("/") 
    os.setsid() 
    os.umask(0) 

    # do second fork
    try: 
        pid = os.fork() 
        if pid > 0:
            # write pid file
            PID = open(pidFile,"w")
            print >>PID,pid
            PID.close()
            # exit from second parent, print eventual PID before
            sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1) 
    
    #close stdin/stdout/stderr
    map(os.close,range(3))

    # write pid file
    PID = open(pidFile,"w")
    print >>PID,pid
    PID.close()

def main(configFile, logFile, pidFile, runAsDaemon=True, verbose=False):
    if runAsDaemon:
        daemonize(pidFile)
    setup_logging(logFile, runAsDaemon, verbose)
    read_config(configFile)

    rejected_location = config.get('logging','rejected-location')
    if rejected_location != None:
        try: 
            os.makedirs(rejected_location);
        except os.error:
            pass

    logger.info('Consumer Startup')
    log_config()
    start()
    stats_loop()

def consume_rejected():
    read_config()
    setup_logging(daemon=False)
    logger.error('consume_rejected function is not implemented yet!')
    pass

    
            
        
        
        
